// Copyright (C) 2020-2023 David Sugar, Tycho Softworks
// This code is licensed under MIT license

#ifndef DRIVERS_HPP_
#define DRIVERS_HPP_

#include "compiler.hpp"
#include "script.hpp"
#include "ipc.hpp"

#include <mutex>
#include <vector>

namespace server {
class Driver {
friend class Session;
friend class script;
friend struct methods;
public:
    using context_t = script::context_t;
    using header_t = script::header_t;
    using check_t = const char *(*)(const script::context_t& scr);
    using line_t = script::line_t;

    struct keyword_t {
        method_t method{nullptr};
        check_t check{nullptr};

        inline keyword_t() = default;
        inline keyword_t(method_t m, check_t c) : method(m), check(c) {}

        inline operator bool() const {
            return method != nullptr;
        }

        inline auto operator!() const {
            return method == nullptr;
        }
    };

    Driver(const Driver&) = delete;
    auto operator=(const Driver&) -> Driver& = delete;

    virtual ~Driver();

    inline auto name() const {
        return name_;
    }

    inline auto limit() const {
        return limit_;
    }

    inline static auto instance() {
        return driver_;
    }

    inline static auto ring_timer() {
        return ring_timer_;
    }

    inline static auto answer_timer() {
        return answer_timer_;
    }


    inline static auto extend_runtime(const std::string& id, const keyword_t& keyword) {
        keywords[id] = keyword;
    }

    inline static auto extend_runtime(const std::initializer_list<std::pair<std::string, keyword_t>>& list) {
        for(auto const& [key, value] : list)
            keywords[key] = value;
    }

    NO_EXPORT static auto config_instance(keyfile& cfg) -> keyfile;
    NO_EXPORT static void reload_instance(const keyfile& cfg);
    NO_EXPORT static void start_instance(const keyfile& ini);
    NO_EXPORT static void drop_instance();
    NO_EXPORT static void stop_instance();

    static auto count() -> size_t;
    static auto get(std::string_view sid) -> session_t&;
    static auto get_script(const std::string& name) -> script&;
    static auto timer(const char *id, int value = 0) -> int;
    static auto attach(const session_t& s) -> bool;
    static void release(const std::string_view& sid);
    static auto control(const std::vector<std::string_view>& list) -> bool;

protected:
    static std::unordered_map<std::string_view, keyword_t> keywords;
    static std::unordered_map<std::string_view, const char *> groups;

    explicit Driver(const std::string& id) noexcept;

    virtual void startup(const keyfile& config) = 0;
    virtual void restart(const keyfile& config);
    virtual void configure(keyfile& config);
    virtual void shutdown();
    virtual void disconnect();

    virtual void drop(Session *session);
    virtual void answer(Session *session);

    // check and coding related functions

    inline static auto keyword(const std::string& key) {
        return keywords[key];
    }

    inline static auto check(check_t handler, const context_t& ctx) {
        return (*(handler))(ctx);
    }

    inline static auto event_group(const char *id) {
        auto result = groups[id];
        return result ? result : id;
    }

    static void notify(bool set);

    unsigned limit_;

    static ipc_system *sys_mapped;
    static std::atomic<bool> online;

private:
    std::string name_;

    static Driver *driver_;
    static unsigned ring_timer_;
    static unsigned answer_timer_;
};

class Plugin {
public:
    Plugin(const Plugin&) = delete;
    auto operator=(const Plugin&) -> Plugin& = delete;

    virtual ~Plugin();

    inline auto name() const {
        return name_;
    }

    virtual auto speak(char *out, size_t size, const char *request, const char *value) -> bool;

    NO_EXPORT static void config(keyfile& cfg);
    NO_EXPORT static void reload(const keyfile& cfg);
    NO_EXPORT static void start(const keyfile& ini);
    NO_EXPORT static void stop();
    NO_EXPORT static auto find(const std::string_view& from) -> Plugin *;
    NO_EXPORT static auto find_locale(std::string_view from) -> Plugin *;
    NO_EXPORT static auto get_formatted(Session *s, const char *keyword) -> const char *;

protected:
    explicit Plugin(const std::string& id) noexcept;

    virtual void startup(const keyfile& config);
    virtual void restart(const keyfile& config);
    virtual void configure(keyfile& config);
    virtual void shutdown();
    virtual auto format(Session *s, const char *keyword) -> const char *;

private:
    friend class Session;

    std::string name_;
    Plugin *next_;

    static Plugin *plugins_;
};

extern std::atomic<bool> running;		// global server run flag
extern bool ipv6;					// global ipv6 setting
} // end namespace

/*!
 * Base class for deriving and managing plugin driver.
 * \file drivers.hpp
 */
#endif
