// Copyright (C) 2020-2023 David Sugar, Tycho Softworks
// This code is licensed under MIT license

#ifndef SYMBOLS_HPP_
#define SYMBOLS_HPP_

#include "compiler.hpp"
#include "memory.hpp"

#include <climits>

namespace server {
class symbols final : private mempager {
public:
    using init_t = std::initializer_list<std::pair<std::string,std::string>>;

    static const unsigned hash_size = TUNE_SYMBOLS_INDEX;
    static const unsigned max_pager = TUNE_SYMBOLS_PAGER;
    static const unsigned max_array = TUNE_SYMBOLS_PAGER / 4;
    static const unsigned short var_size = TUNE_SYMBOL_SIZE;

    symbols();
    ~symbols() final;

    inline auto dup(const char *cp) {
        return mempager::dup(cp);
    }

    inline auto alloc(size_t size) {
        return mempager::alloc(size);
    }

    auto exists(const std::string_view& id) const -> bool;
    auto get_value(const std::string_view& id) const -> const char *;              // op-after
    auto from_value(const std::string_view& id, short offset) const -> const char *;    // no-op
    auto get_count(const std::string_view& id) const -> unsigned;
    auto get_size(const std::string_view& id) const-> size_t;
    auto set_value(const std::string_view& id, const char *value) -> bool;
    auto set_exists(const std::string_view& id, std::string_view value) -> bool;

    auto get_storage(const std::string_view& id) -> char *;
    auto clear_symbol(const std::string_view& id) -> bool;
    auto make_const(const std::string_view& id, const char *value) -> bool;
    auto make_value(const std::string_view& id, unsigned short size = var_size) -> bool;
    auto make_counter(const std::string_view& id, unsigned short max) -> bool;
    auto make_storage(const std::string_view& id, unsigned short size = var_size, const char *cval = nullptr) -> char *;
    auto make_const(const init_t& list) -> unsigned;
    void purge();

    inline static auto to_short(const std::string& s, short value = 0) {
        return static_cast<short>(to_value(s, value));
    }

    inline static auto to_size(const std::string& s) -> unsigned short {
        return static_cast<unsigned short>(to_value(s));
    }

    static auto is_number(const char *cp) -> bool;
    static auto is_number(const std::string& s) -> bool;
    static auto is_path(const char *cp) -> bool;
    static auto is_path(const std::string& s) -> bool;
    static auto is_symbol(const char *cp) -> bool;
    static auto is_symbol(const std::string& s) -> bool;
    static auto to_value(const std::string& s, int value = 0, int min = INT_MIN, int max = INT_MAX) -> int;

private:
    struct symbol_t {
        symbol_t *next;
        char *data; // const fixed, sequence source...
        enum { NORMAL, CONST, SEQUENCE, COUNT, ARRAY, FIFO, LIFO } type;
        unsigned short size, count; // size = total or entity, count = entities
        union {
            char buffer[4];
            struct {
                unsigned short head, tail;
            };
            struct {      // sequence or array positioning...
                unsigned short pos;
            };
        };
        char id[1];
    };

    symbol_t *index[hash_size];

    auto find_symbol(const std::string_view& id) const -> symbol_t *;
    auto make_symbol(std::string_view id) -> symbol_t *;
};
} // end namespace

/*!
 * Common symbol pool related functions and extensions.
 * \file symbols.hpp
 */
#endif
